@@ -11,6 +11,27 @@ import ExprNodes
 from ModuleNode import ModuleNode
 from Errors import warning, error, InternalError
 
+
+class Ctx(object):
+    #  Parsing context
+    level = 'other'
+    visibility = 'private'
+    cdef_flag = 0
+    typedef_flag = 0
+    api = 0
+    nogil = 0
+    
+    def __init__(self, **kwds):
+        self.__dict__.update(kwds)
+    
+    def __call__(self, **kwds):
+        ctx = Ctx()
+        d = ctx.__dict__
+        d.update(self.__dict__)
+        d.update(kwds)
+        return ctx
+
+
 def p_ident(s, message = "Expected an identifier"):
     if s.sy == 'IDENT':
         name = s.systring
@@ -36,9 +57,7 @@ def p_ident_list(s):
 #------------------------------------------
 
 def p_binop_expr(s, ops, p_sub_expr):
-    #print "p_binop_expr:", ops, p_sub_expr ###
     n1 = p_sub_expr(s)
-    #print "p_binop_expr(%s):" % p_sub_expr, s.sy ###
     while s.sy in ops:
         op = s.sy
         pos = s.position()
@@ -50,7 +69,6 @@ def p_binop_expr(s, ops, p_sub_expr):
 #test: and_test ('or' and_test)* | lambdef
 
 def p_simple_expr(s):
-    #return p_binop_expr(s, ('or',), p_and_test)
     return p_rassoc_binop_expr(s, ('or',), p_and_test)
 
 def p_rassoc_binop_expr(s, ops, p_subexpr):
@@ -669,18 +687,24 @@ def p_global_statement(s):
     names = p_ident_list(s)
     return Nodes.GlobalNode(pos, names = names)
 
+inplace_operators = ('+=', '-=', '*=', '/=', '%=', '**=',
+    '<<=', '>>=', '&=', '^=', '|=')
+
 def p_expression_or_assignment(s):
-    expr_list = [p_expr(s)]
-    while s.sy == '=':
-        s.next()
-        expr_list.append(p_expr(s))
-    if len(expr_list) == 1:
-        expr = expr_list[0]
+    pos = s.position()
+    expr = p_expr(s)
+    if s.sy in inplace_operators:
+        return p_inplace_operation(s, expr)
+    elif s.sy <> '=':
         if isinstance(expr, ExprNodes.StringNode):
             return Nodes.PassStatNode(expr.pos)
         else:
             return Nodes.ExprStatNode(expr.pos, expr = expr)
     else:
+        expr_list = [expr]
+        while s.sy == '=':
+            s.next()
+            expr_list.append(p_expr(s))
         expr_list_list = []
         flatten_parallel_assignments(expr_list, expr_list_list)
         nodes = []
@@ -699,6 +723,13 @@ def p_expression_or_assignment(s):
         else:
             return Nodes.ParallelAssignmentNode(nodes[0].pos, stats = nodes)
 
+def p_inplace_operation(s, lhs):
+    pos = s.position()
+    op = s.sy
+    s.next()
+    rhs = p_expr(s)
+    return Nodes.AugmentedAssignmentNode(pos, lhs = lhs, operator = op, rhs = rhs)
+
 def flatten_parallel_assignments(input, output):
     #  The input is a list of expression nodes, representing 
     #  the LHSs and RHS of one (possibly cascaded) assignment 
@@ -842,7 +873,7 @@ def p_import_statement(s):
         stats.append(stat)
     return Nodes.StatListNode(pos, stats = stats)
 
-def p_from_import_statement(s, level):
+def p_from_import_statement(s, ctx):
     # s.sy == 'from'
     pos = s.position()
     s.next()
@@ -853,17 +884,18 @@ def p_from_import_statement(s, level):
         s.next()
     else:
         s.error("Expected 'import' or 'cimport'")
-    if kind == 'cimport' and level not in ('module', 'module_pxd'):
+    if kind == 'cimport' and ctx.level not in ('module', 'module_pxd'):
         s.error("cimport statement not allowed in this context")
     if s.sy == '*':
         s.error("'import *' not supported")
-    imported_names = [p_imported_name(s)]
+    is_cimport = kind == 'cimport'
+    imported_names = [p_imported_name(s, is_cimport)]
     while s.sy == ',':
         s.next()
-        imported_names.append(p_imported_name(s))
+        imported_names.append(p_imported_name(s, is_cimport))
     if kind == 'cimport':
-        for (name_pos, name, as_name) in imported_names:
-            local_name = as_name or name
+        for imp in imported_names:
+            local_name = imp.as_name or imp.name
             s.add_type_name(local_name)
         return Nodes.FromCImportStatNode(pos,
             module_name = dotted_name,
@@ -871,15 +903,15 @@ def p_from_import_statement(s, level):
     else:
         imported_name_strings = []
         items = []
-        for (name_pos, name, as_name) in imported_names:
+        for imp in imported_names:
             imported_name_strings.append(
-                ExprNodes.StringNode(name_pos, value = name))
+                ExprNodes.StringNode(imp.pos, value = imp.name))
             items.append(
-                (name,
-                 ExprNodes.NameNode(name_pos, 
-                 	name = as_name or name)))
+                (imp.name,
+                 ExprNodes.NameNode(imp.pos, 
+                 	name = imp.as_name or imp.name)))
         import_list = ExprNodes.ListNode(
-            imported_names[0][0], args = imported_name_strings)
+            imported_names[0].pos, args = imported_name_strings)
         return Nodes.FromImportStatNode(pos,
             module = ExprNodes.ImportNode(dotted_name_pos,
                 module_name = ExprNodes.StringNode(dotted_name_pos,
@@ -887,11 +919,29 @@ def p_from_import_statement(s, level):
                 name_list = import_list),
             items = items)
 
-def p_imported_name(s):
+class ImportedName(object):
+    #  pos
+    #  name
+    #  as_name
+    #  kind     'class', 'struct', 'union', None
+    
+    def __init__(self, pos, name, as_name, kind):
+        self.pos = pos
+        self.name = name
+        self.as_name = as_name
+        self.kind = kind
+
+imported_name_kinds = ('class', 'struct', 'union')
+
+def p_imported_name(s, is_cimport):
     pos = s.position()
+    kind = None
+    if is_cimport and s.systring in imported_name_kinds:
+        kind = s.systring
+        s.next()
     name = p_ident(s)
     as_name = p_as_name(s)
-    return (pos, name, as_name)
+    return ImportedName(pos, name, as_name, kind)
 
 def p_dotted_name(s, as_allowed):
     pos = s.position()
@@ -1042,50 +1092,6 @@ def p_old_style_integer_for_statement(s, target):
         body = body,
         else_clause = else_clause)
 
-#def p_for_statement(s):
-#	# s.sy == 'for'
-#	pos = s.position()
-#	s.next()
-#	target = p_for_target(s)
-#	if s.sy == 'in':
-#		s.next()
-#		iterator = p_for_iterator(s)
-#		body = p_suite(s)
-#		else_clause = p_else_clause(s)
-#		return Nodes.ForInStatNode(pos, 
-#			target = target,
-#			iterator = iterator,
-#			body = body,
-#			else_clause = else_clause)
-#	elif s.sy == 'from':
-#		s.next()
-#		bound1 = p_bit_expr(s)
-#		rel1 = p_for_from_relation(s)
-#		name2_pos = s.position()
-#		name2 = p_ident(s)
-#		rel2_pos = s.position()
-#		rel2 = p_for_from_relation(s)
-#		bound2 = p_bit_expr(s)
-#		if not target.is_name:
-#			error(target.pos, 
-#				"Target of for-from statement must be a variable name")
-#		elif name2 <> target.name:
-#			error(name2_pos,
-#				"Variable name in for-from range does not match target")
-#		if rel1[0] <> rel2[0]:
-#			error(rel2_pos,
-#				"Relation directions in for-from do not match")
-#		body = p_suite(s)
-#		else_clause = p_else_clause(s)
-#		return Nodes.ForFromStatNode(pos,
-#			target = target,
-#			bound1 = bound1,
-#			relation1 = rel1,
-#			relation2 = rel2,
-#			bound2 = bound2,
-#			body = body,
-#			else_clause = else_clause)
-
 def p_inequality_relation(s):
     if s.sy in inequality_relations:
         op = s.sy
@@ -1156,7 +1162,7 @@ def p_except_clause(s):
     return Nodes.ExceptClauseNode(pos,
         pattern = exc_type, target = exc_value, body = body)
 
-def p_include_statement(s, level):
+def p_include_statement(s, ctx):
     pos = s.position()
     s.next() # 'include'
     _, include_file_name = p_string_literal(s)
@@ -1164,10 +1170,11 @@ def p_include_statement(s, level):
     if s.compile_time_eval:
         include_file_path = s.context.find_include_file(include_file_name, pos)
         if include_file_path:
+            s.included_files.append(include_file_name)
             f = open(include_file_path, "rU")
-            s2 = PyrexScanner(f, include_file_path, s)
+            s2 = PyrexScanner(f, include_file_path, parent_scanner = s)
             try:
-                tree = p_statement_list(s2, level)
+                tree = p_statement_list(s2, ctx)
             finally:
                 f.close()
             return tree
@@ -1188,7 +1195,7 @@ def p_with_statement(s):
     else:
         s.error("Only 'with nogil' implemented")
     
-def p_simple_statement(s, level):
+def p_simple_statement(s, ctx):
     if s.sy == 'global':
         node = p_global_statement(s)
     elif s.sy == 'print':
@@ -1204,13 +1211,13 @@ def p_simple_statement(s, level):
     elif s.sy == 'raise':
         node = p_raise_statement(s)
     elif s.sy == 'cimport':
-        if level not in ('module', 'module_pxd'):
+        if ctx.level not in ('module', 'module_pxd'):
             s.error("cimport statement not allowed in this context")
         node = p_import_statement(s)
     elif s.sy == 'import':
         node = p_import_statement(s)
     elif s.sy == 'from':
-        node = p_from_import_statement(s, level)
+        node = p_from_import_statement(s, ctx)
     elif s.sy == 'assert':
         node = p_assert_statement(s)
     elif s.sy == 'pass':
@@ -1219,17 +1226,17 @@ def p_simple_statement(s, level):
         node = p_expression_or_assignment(s)
     return node
 
-def p_simple_statement_list(s, level):
+def p_simple_statement_list(s, ctx):
     # Parse a series of simple statements on one line
     # separated by semicolons.
-    stat = p_simple_statement(s, level)
+    stat = p_simple_statement(s, ctx)
     if s.sy == ';':
         stats = [stat]
         while s.sy == ';':
             s.next()
             if s.sy in ('NEWLINE', 'EOF'):
                 break
-            stats.append(p_simple_statement(s, level))
+            stats.append(p_simple_statement(s, ctx))
         stat = Nodes.StatListNode(stats[0].pos, stats = stats)
     s.expect_newline("Syntax error in simple statement list")
     return stat
@@ -1254,7 +1261,7 @@ def p_DEF_statement(s):
     s.expect_newline()
     return Nodes.PassStatNode(pos)
 
-def p_IF_statement(s, level, cdef_flag, visibility, api):
+def p_IF_statement(s, ctx):
     pos = s.position()
     saved_eval = s.compile_time_eval
     current_eval = saved_eval
@@ -1264,7 +1271,7 @@ def p_IF_statement(s, level, cdef_flag, visibility, api):
         s.next() # 'IF' or 'ELIF'
         expr = p_compile_time_expr(s)
         s.compile_time_eval = current_eval and bool(expr.compile_time_value(denv))
-        body = p_suite(s, level, cdef_flag, visibility, api = api)
+        body = p_suite(s, ctx)
         if s.compile_time_eval:
             result = body
             current_eval = 0
@@ -1273,7 +1280,7 @@ def p_IF_statement(s, level, cdef_flag, visibility, api):
     if s.sy == 'ELSE':
         s.next()
         s.compile_time_eval = current_eval
-        body = p_suite(s, level, cdef_flag, visibility, api = api)
+        body = p_suite(s, ctx)
         if current_eval:
             result = body
     if not result:
@@ -1281,46 +1288,47 @@ def p_IF_statement(s, level, cdef_flag, visibility, api):
     s.compile_time_eval = saved_eval
     return result
 
-def p_statement(s, level, cdef_flag = 0, visibility = 'private', api = 0):
+def p_statement(s, ctx):
+    cdef_flag = ctx.cdef_flag
     if s.sy == 'ctypedef':
-        if level not in ('module', 'module_pxd'):
+        if ctx.level not in ('module', 'module_pxd'):
             s.error("ctypedef statement not allowed here")
-        if api:
-            error(s.pos, "'api' not allowed with 'ctypedef'")
-        return p_ctypedef_statement(s, level, visibility, api)
+        #if ctx.api:
+        #	error(s.pos, "'api' not allowed with 'ctypedef'")
+        return p_ctypedef_statement(s, ctx)
     elif s.sy == 'DEF':
         return p_DEF_statement(s)
     elif s.sy == 'IF':
-        return p_IF_statement(s, level, cdef_flag, visibility, api)
+        return p_IF_statement(s, ctx)
     else:
         if s.sy == 'cdef':
             cdef_flag = 1
             s.next()
         if cdef_flag:
-            if level not in ('module', 'module_pxd', 'function', 'c_class', 'c_class_pxd'):
+            if ctx.level not in ('module', 'module_pxd', 'function', 'c_class', 'c_class_pxd'):
                 s.error('cdef statement not allowed here')
-            return p_cdef_statement(s, level, visibility, api)
+            return p_cdef_statement(s, ctx)
         else:
-            if api:
+            if ctx.api:
                 error(s.pos, "'api' not allowed with this statement")
             if s.sy == 'def':
-                if level not in ('module', 'class', 'c_class', 'property'):
+                if ctx.level not in ('module', 'class', 'c_class', 'property'):
                     s.error('def statement not allowed here')
                 return p_def_statement(s)
             elif s.sy == 'class':
-                if level <> 'module':
+                if ctx.level <> 'module':
                     s.error("class definition not allowed here")
                 return p_class_statement(s)
             elif s.sy == 'include':
-                #if level not in ('module', 'module_pxd'):
+                #if ctx.level not in ('module', 'module_pxd'):
                 #	s.error("include statement not allowed here")
-                return p_include_statement(s, level)
-            elif level == 'c_class' and s.sy == 'IDENT' and s.systring == 'property':
+                return p_include_statement(s, ctx)
+            elif ctx.level == 'c_class' and s.sy == 'IDENT' and s.systring == 'property':
                 return p_property_decl(s)
-            elif s.sy == 'pass' and level <> 'property':
+            elif s.sy == 'pass' and ctx.level <> 'property':
                 return p_pass_statement(s, with_newline = 1)
             else:
-                if level in ('c_class', 'c_class_pxd', 'property'):
+                if ctx.level in ('c_class', 'c_class_pxd', 'property'):
                     s.error("Executable statement not allowed here")
                 if s.sy == 'if':
                     return p_if_statement(s)
@@ -1333,23 +1341,20 @@ def p_statement(s, level, cdef_flag = 0, visibility = 'private', api = 0):
                 elif s.sy == 'with':
                     return p_with_statement(s)
                 else:
-                    return p_simple_statement_list(s, level)
+                    return p_simple_statement_list(s, ctx)
 
-def p_statement_list(s, level,
-        cdef_flag = 0, visibility = 'private', api = 0):
+def p_statement_list(s, ctx):
     # Parse a series of statements separated by newlines.
     pos = s.position()
     stats = []
     while s.sy not in ('DEDENT', 'EOF'):
-        stats.append(p_statement(s, level,
-            cdef_flag = cdef_flag, visibility = visibility, api = api))
+        stats.append(p_statement(s, ctx))
     if len(stats) == 1:
         return stats[0]
     else:
         return Nodes.StatListNode(pos, stats = stats)
 
-def p_suite(s, level = 'other', cdef_flag = 0,
-        visibility = 'private', with_doc = 0, with_pseudo_doc = 0, api = 0):
+def p_suite(s, ctx = Ctx(), with_doc = 0, with_pseudo_doc = 0):
     pos = s.position()
     s.expect(':')
     doc = None
@@ -1359,17 +1364,13 @@ def p_suite(s, level = 'other', cdef_flag = 0,
         s.expect_indent()
         if with_doc or with_pseudo_doc:
             doc = p_doc_string(s)
-        body = p_statement_list(s, 
-            level = level,
-            cdef_flag = cdef_flag, 
-            visibility = visibility,
-            api = api)
+        body = p_statement_list(s, ctx)
         s.expect_dedent()
     else:
-        if api:
+        if ctx.api:
             error(s.pos, "'api' not allowed with this statement")
-        if level in ('module', 'class', 'function', 'other'):
-            body = p_simple_statement_list(s, level)
+        if ctx.level in ('module', 'class', 'function', 'other'):
+            body = p_simple_statement_list(s, ctx)
         else:
             body = p_pass_statement(s)
             s.expect_newline("Syntax error in declarations")
@@ -1485,7 +1486,7 @@ def p_opt_cname(s):
         cname = None
     return cname
 
-def p_c_declarator(s, empty = 0, is_type = 0, cmethod_flag = 0, nonempty = 0,
+def p_c_declarator(s, ctx = Ctx(), empty = 0, is_type = 0, cmethod_flag = 0, nonempty = 0,
         calling_convention_allowed = 0):
     # If empty is true, the declarator must be empty. If nonempty is true,
     # the declarator must be nonempty. Otherwise we don't care.
@@ -1496,13 +1497,14 @@ def p_c_declarator(s, empty = 0, is_type = 0, cmethod_flag = 0, nonempty = 0,
         s.next()
         if s.sy == ')' or looking_at_type(s):
             base = Nodes.CNameDeclaratorNode(pos, name = "", cname = None)
-            result = p_c_func_declarator(s, pos, base, cmethod_flag)
+            result = p_c_func_declarator(s, pos, ctx, base, cmethod_flag)
         else:
-            result = p_c_declarator(s, empty, is_type, cmethod_flag, nonempty,
+            result = p_c_declarator(s, ctx, empty = empty, is_type = is_type,
+                cmethod_flag = cmethod_flag, nonempty = nonempty,
                 calling_convention_allowed = 1)
             s.expect(')')
     else:
-        result = p_c_simple_declarator(s, empty, is_type, cmethod_flag, nonempty)
+        result = p_c_simple_declarator(s, ctx, empty, is_type, cmethod_flag, nonempty)
     if not calling_convention_allowed and result.calling_convention and s.sy <> '(':
         error(s.position(), "%s on something that is not a function"
             % result.calling_convention)
@@ -1512,7 +1514,7 @@ def p_c_declarator(s, empty = 0, is_type = 0, cmethod_flag = 0, nonempty = 0,
             result = p_c_array_declarator(s, result)
         else: # sy == '('
             s.next()
-            result = p_c_func_declarator(s, pos, result, cmethod_flag)
+            result = p_c_func_declarator(s, pos, ctx, result, cmethod_flag)
         cmethod_flag = 0
     return result
 
@@ -1526,9 +1528,9 @@ def p_c_array_declarator(s, base):
     s.expect(']')
     return Nodes.CArrayDeclaratorNode(pos, base = base, dimension = dim)
 
-def p_c_func_declarator(s, pos, base, cmethod_flag):
+def p_c_func_declarator(s, pos, ctx, base, cmethod_flag):
     #  Opening paren has already been skipped
-    args = p_c_arg_list(s, in_pyfunc = 0, cmethod_flag = cmethod_flag,
+    args = p_c_arg_list(s, ctx, cmethod_flag = cmethod_flag,
         nonempty_declarators = 0)
     ellipsis = p_optional_ellipsis(s)
     s.expect(')')
@@ -1538,19 +1540,21 @@ def p_c_func_declarator(s, pos, base, cmethod_flag):
     return Nodes.CFuncDeclaratorNode(pos, 
         base = base, args = args, has_varargs = ellipsis,
         exception_value = exc_val, exception_check = exc_check,
-        nogil = nogil or with_gil, with_gil = with_gil)
+        nogil = nogil or ctx.nogil or with_gil, with_gil = with_gil)
 
-def p_c_simple_declarator(s, empty, is_type, cmethod_flag, nonempty):
+def p_c_simple_declarator(s, ctx, empty, is_type, cmethod_flag, nonempty):
     pos = s.position()
     calling_convention = p_calling_convention(s)
     if s.sy == '*':
         s.next()
-        base = p_c_declarator(s, empty, is_type, cmethod_flag, nonempty)
+        base = p_c_declarator(s, ctx, empty = empty, is_type = is_type,
+            cmethod_flag = cmethod_flag, nonempty = nonempty)
         result = Nodes.CPtrDeclaratorNode(pos, 
             base = base)
     elif s.sy == '**': # scanner returns this as a single token
         s.next()
-        base = p_c_declarator(s, empty, is_type, cmethod_flag, nonempty)
+        base = p_c_declarator(s, ctx, empty = empty, is_type = is_type,
+            cmethod_flag = cmethod_flag, nonempty = nonempty)
         result = Nodes.CPtrDeclaratorNode(pos,
             base = Nodes.CPtrDeclaratorNode(pos,
                 base = base))
@@ -1605,28 +1609,14 @@ def p_exception_value_clause(s):
 
 c_arg_list_terminators = ('*', '**', '.', ')')
 
-#def p_c_arg_list(s, in_pyfunc, cmethod_flag = 0, nonempty_declarators = 0,
-#		kw_only = 0):
-#	args = []
-#	if s.sy not in c_arg_list_terminators:
-#		args.append(p_c_arg_decl(s, in_pyfunc, cmethod_flag,
-#			nonempty = nonempty_declarators, kw_only = kw_only))
-#		while s.sy == ',':
-#			s.next()
-#			if s.sy in c_arg_list_terminators:
-#				break
-#			args.append(p_c_arg_decl(s, in_pyfunc), nonempty = nonempty_declarators,
-#				kw_only = kw_only)
-#	return args
-
-def p_c_arg_list(s, in_pyfunc, cmethod_flag = 0, nonempty_declarators = 0,
+def p_c_arg_list(s, ctx = Ctx(), in_pyfunc = 0, cmethod_flag = 0, nonempty_declarators = 0,
         kw_only = 0):
     #  Comma-separated list of C argument declarations, possibly empty.
     #  May have a trailing comma.
     args = []
     is_self_arg = cmethod_flag
     while s.sy not in c_arg_list_terminators:
-        args.append(p_c_arg_decl(s, in_pyfunc, is_self_arg,
+        args.append(p_c_arg_decl(s, ctx, in_pyfunc, is_self_arg,
             nonempty = nonempty_declarators, kw_only = kw_only))
         if s.sy <> ',':
             break
@@ -1641,12 +1631,12 @@ def p_optional_ellipsis(s):
     else:
         return 0
 
-def p_c_arg_decl(s, in_pyfunc, cmethod_flag = 0, nonempty = 0, kw_only = 0):
+def p_c_arg_decl(s, ctx, in_pyfunc, cmethod_flag = 0, nonempty = 0, kw_only = 0):
     pos = s.position()
     not_none = 0
     default = None
     base_type = p_c_base_type(s, cmethod_flag)
-    declarator = p_c_declarator(s, nonempty = nonempty)
+    declarator = p_c_declarator(s, ctx, nonempty = nonempty)
     if s.sy == 'not':
         s.next()
         if s.sy == 'IDENT' and s.systring == 'None':
@@ -1673,52 +1663,58 @@ def p_api(s):
     else:
         return 0
 
-def p_cdef_statement(s, level, visibility = 'private', api = 0):
+def p_cdef_statement(s, ctx):
+    ctx = ctx(cdef_flag = 1)
     pos = s.position()
-    visibility = p_visibility(s, visibility)
-    api = api or p_api(s)
-    if api:
-        if visibility not in ('private', 'public'):
+    ctx.visibility = p_visibility(s, ctx.visibility)
+    ctx.api = ctx.api or p_api(s)
+    if ctx.api:
+        if ctx.visibility not in ('private', 'public'):
             error(pos, "Cannot combine 'api' with '%s'" % visibility)
-    if visibility == 'extern' and s.sy == 'from':
-            return p_cdef_extern_block(s, level, pos)
-    elif s.sy == ':':
-        return p_cdef_block(s, level, visibility, api)
+    if ctx.visibility == 'extern' and s.sy == 'from':
+        return p_cdef_extern_block(s, pos, ctx)
+    if p_nogil(s):
+        ctx.nogil = 1
+    if s.sy == ':':
+        return p_cdef_block(s, ctx)
     elif s.sy == 'class':
-        if level not in ('module', 'module_pxd'):
+        if ctx.level not in ('module', 'module_pxd'):
             error(pos, "Extension type definition not allowed here")
         #if api:
         #	error(pos, "'api' not allowed with extension class")
-        return p_c_class_definition(s, level, pos, visibility = visibility, api = api)
+        return p_c_class_definition(s, pos, ctx)
     elif s.sy == 'IDENT' and s.systring in struct_union_or_enum:
-        if level not in ('module', 'module_pxd'):
+        if ctx.level not in ('module', 'module_pxd'):
             error(pos, "C struct/union/enum definition not allowed here")
-        #if visibility == 'public':
+        #if ctx.visibility == 'public':
         #	error(pos, "Public struct/union/enum definition not implemented")
-        #if api:
+        #if ctx.api:
         #	error(pos, "'api' not allowed with '%s'" % s.systring)
         if s.systring == "enum":
-            return p_c_enum_definition(s, pos, level, visibility)
+            return p_c_enum_definition(s, pos, ctx)
         else:
-            return p_c_struct_or_union_definition(s, pos, level, visibility)
+            return p_c_struct_or_union_definition(s, pos, ctx)
     elif s.sy == 'pass':
         node = p_pass_statement(s)
         s.expect_newline('Expected a newline')
         return node
     else:
-        return p_c_func_or_var_declaration(s, level, pos, visibility, api)
+        return p_c_func_or_var_declaration(s, pos, ctx)
 
-def p_cdef_block(s, level, visibility, api):
-    return p_suite(s, level, cdef_flag = 1, visibility = visibility, api = api)
+def p_cdef_block(s, ctx):
+    return p_suite(s, ctx(cdef_flag = 1))
 
-def p_cdef_extern_block(s, level, pos):
+def p_cdef_extern_block(s, pos, ctx):
     include_file = None
     s.expect('from')
     if s.sy == '*':
         s.next()
     else:
         _, include_file = p_string_literal(s)
-    body = p_suite(s, level, cdef_flag = 1, visibility = 'extern')
+    ctx = ctx(cdef_flag = 1, visibility = 'extern')
+    if p_nogil(s):
+        ctx.nogil = 1
+    body = p_suite(s, ctx)
     return Nodes.CDefExternNode(pos,
         include_file = include_file,
         body = body)
@@ -1727,7 +1723,7 @@ struct_union_or_enum = (
     "struct", "union", "enum"
 )
 
-def p_c_enum_definition(s, pos, level, visibility, typedef_flag = 0):
+def p_c_enum_definition(s, pos, ctx):
     # s.sy == ident 'enum'
     s.next()
     if s.sy == 'IDENT':
@@ -1749,9 +1745,10 @@ def p_c_enum_definition(s, pos, level, visibility, typedef_flag = 0):
         while s.sy not in ('DEDENT', 'EOF'):
             p_c_enum_line(s, items)
         s.expect_dedent()
-    return Nodes.CEnumDefNode(pos, name = name, cname = cname,
-        items = items, typedef_flag = typedef_flag, visibility = visibility,
-        in_pxd = level == 'module_pxd')
+    return Nodes.CEnumDefNode(pos, name = name, cname = cname, items = items,
+        typedef_flag = ctx.typedef_flag,
+        visibility = ctx.visibility,
+        in_pxd = ctx.level == 'module_pxd')
 
 def p_c_enum_line(s, items):
     if s.sy <> 'pass':
@@ -1776,11 +1773,11 @@ def p_c_enum_item(s, items):
     items.append(Nodes.CEnumDefItemNode(pos, 
         name = name, cname = cname, value = value))
 
-def p_c_struct_or_union_definition(s, pos, level, visibility, typedef_flag = 0):
+def p_c_struct_or_union_definition(s, pos, ctx):
     # s.sy == ident 'struct' or 'union'
     kind = s.systring
     s.next()
-    name = p_ident(s)
+    module_path, name = p_qualified_name(s)
     cname = p_opt_cname(s)
     s.add_type_name(name)
     attributes = None
@@ -1789,10 +1786,11 @@ def p_c_struct_or_union_definition(s, pos, level, visibility, typedef_flag = 0):
         s.expect('NEWLINE')
         s.expect_indent()
         attributes = []
+        body_ctx = Ctx()
         while s.sy <> 'DEDENT':
             if s.sy <> 'pass':
-                attributes.append(
-                    p_c_func_or_var_declaration(s, level = 'other', pos = s.position()))
+                attributes.append(p_c_func_or_var_declaration(s,
+                    s.position(), body_ctx))
             else:
                 s.next()
                 s.expect_newline("Expected a newline")
@@ -1800,9 +1798,11 @@ def p_c_struct_or_union_definition(s, pos, level, visibility, typedef_flag = 0):
     else:
         s.expect_newline("Syntax error in struct or union definition")
     return Nodes.CStructOrUnionDefNode(pos, 
-        name = name, cname = cname, kind = kind, attributes = attributes,
-        typedef_flag = typedef_flag, visibility = visibility,
-        in_pxd = level == 'module_pxd')
+        name = name, cname = cname, module_path = module_path,
+        kind = kind, attributes = attributes,
+        typedef_flag = ctx.typedef_flag,
+        visibility = ctx.visibility,
+        in_pxd = ctx.level == 'module_pxd')
 
 def p_visibility(s, prev_visibility):
     pos = s.position()
@@ -1815,20 +1815,20 @@ def p_visibility(s, prev_visibility):
         s.next()
     return visibility
 
-def p_c_func_or_var_declaration(s, level, pos, visibility = 'private', api = 0):
-    cmethod_flag = level in ('c_class', 'c_class_pxd')
+def p_c_func_or_var_declaration(s, pos, ctx):
+    cmethod_flag = ctx.level in ('c_class', 'c_class_pxd')
     base_type = p_c_base_type(s)
-    declarator = p_c_declarator(s, cmethod_flag = cmethod_flag, nonempty = 1)
+    declarator = p_c_declarator(s, ctx, cmethod_flag = cmethod_flag, nonempty = 1)
     if s.sy == ':':
-        if level not in ('module', 'c_class'):
+        if ctx.level not in ('module', 'c_class'):
             s.error("C function definition not allowed here")
-        suite = p_suite(s, 'function', with_pseudo_doc = 1)
+        suite = p_suite(s, Ctx(level = 'function'), with_pseudo_doc = 1)
         result = Nodes.CFuncDefNode(pos,
-            visibility = visibility,
+            visibility = ctx.visibility,
             base_type = base_type,
             declarator = declarator, 
             body = suite,
-            api = api)
+            api = ctx.api)
     else:
         #if api:
         #	error(pos, "'api' not allowed with variable declaration")
@@ -1837,38 +1837,41 @@ def p_c_func_or_var_declaration(s, level, pos, visibility = 'private', api = 0):
             s.next()
             if s.sy == 'NEWLINE':
                 break
-            declarator = p_c_declarator(s, cmethod_flag = cmethod_flag, nonempty = 1)
+            declarator = p_c_declarator(s, ctx, cmethod_flag = cmethod_flag, nonempty = 1)
             declarators.append(declarator)
         s.expect_newline("Syntax error in C variable declaration")
         result = Nodes.CVarDefNode(pos, 
-            visibility = visibility,
+            visibility = ctx.visibility,
             base_type = base_type, 
             declarators = declarators,
-            in_pxd = level == 'module_pxd',
-            api = api)
+            in_pxd = ctx.level == 'module_pxd',
+            api = ctx.api)
     return result
 
-def p_ctypedef_statement(s, level, visibility = 'private', api = 0):
+def p_ctypedef_statement(s, ctx):
     # s.sy == 'ctypedef'
     pos = s.position()
     s.next()
-    visibility = p_visibility(s, visibility)
+    visibility = p_visibility(s, ctx.visibility)
+    api = p_api(s)
+    ctx = ctx(typedef_flag = 1, visibility = visibility)
+    if api:
+        ctx.api = 1
     if s.sy == 'class':
-        return p_c_class_definition(s, level, pos,
-            visibility = visibility, typedef_flag = 1, api = api)
+        return p_c_class_definition(s, pos, ctx)
     elif s.sy == 'IDENT' and s.systring in ('struct', 'union', 'enum'):
         if s.systring == 'enum':
-            return p_c_enum_definition(s, pos, level, visibility, typedef_flag = 1)
+            return p_c_enum_definition(s, pos, ctx)
         else:
-            return p_c_struct_or_union_definition(s, pos, level, visibility,
-                typedef_flag = 1)
+            return p_c_struct_or_union_definition(s, pos, ctx)
     else:
         base_type = p_c_base_type(s)
-        declarator = p_c_declarator(s, is_type = 1, nonempty = 1)
+        declarator = p_c_declarator(s, ctx, is_type = 1, nonempty = 1)
         s.expect_newline("Syntax error in ctypedef statement")
         return Nodes.CTypeDefNode(pos,
-            base_type = base_type, declarator = declarator, visibility = visibility,
-            in_pxd = level == 'module_pxd')
+            base_type = base_type, declarator = declarator,
+            visibility = ctx.visibility,
+            in_pxd = ctx.level == 'module_pxd')
 
 def p_def_statement(s):
     # s.sy == 'def'
@@ -1896,7 +1899,7 @@ def p_def_statement(s):
     s.expect(')')
     if p_nogil(s):
         error(s.pos, "Python function cannot be declared nogil")
-    doc, body = p_suite(s, 'function', with_doc = 1)
+    doc, body = p_suite(s, Ctx(level = 'function'), with_doc = 1)
     return Nodes.DefNode(pos, name = name, args = args, 
         star_arg = star_arg, starstar_arg = starstar_arg,
         doc = doc, body = body)
@@ -1917,24 +1920,25 @@ def p_class_statement(s):
         s.expect(')')
     else:
         base_list = []
-    doc, body = p_suite(s, 'class', with_doc = 1)
+    doc, body = p_suite(s, Ctx(level = 'class'), with_doc = 1)
     return Nodes.PyClassDefNode(pos,
         name = class_name,
         bases = ExprNodes.TupleNode(pos, args = base_list),
         doc = doc, body = body)
 
-def p_c_class_definition(s, level, pos, 
-        visibility = 'private', typedef_flag = 0, api = 0):
-    # s.sy == 'class'
-    s.next()
-    module_path = []
-    class_name = p_ident(s)
+def p_qualified_name(s):
+    path = []
+    name = p_ident(s)
     while s.sy == '.':
         s.next()
-        module_path.append(class_name)
-        class_name = p_ident(s)
-    if module_path and visibility <> 'extern':
-        error(pos, "Qualified class name only allowed for 'extern' C class")
+        path.append(name)
+        name = p_ident(s)
+    return path, name
+
+def p_c_class_definition(s, pos, ctx):
+    # s.sy == 'class'
+    s.next()
+    module_path, class_name = p_qualified_name(s)
     if module_path and s.sy == 'IDENT' and s.systring == 'as':
         s.next()
         as_name = p_ident(s)
@@ -1947,46 +1951,42 @@ def p_c_class_definition(s, level, pos,
     base_class_name = None
     if s.sy == '(':
         s.next()
-        base_class_path = [p_ident(s)]
-        while s.sy == '.':
-            s.next()
-            base_class_path.append(p_ident(s))
+        base_class_path, base_class_name = p_qualified_name(s)
         if s.sy == ',':
             s.error("C class may only have one base class")
         s.expect(')')
-        base_class_module = ".".join(base_class_path[:-1])
-        base_class_name = base_class_path[-1]
+        base_class_module = ".".join(base_class_path)
     if s.sy == '[':
-        if visibility not in ('public', 'extern'):
+        if ctx.visibility not in ('public', 'extern'):
             error(s.position(), "Name options only allowed for 'public' or 'extern' C class")
         objstruct_name, typeobj_name = p_c_class_options(s)
     if s.sy == ':':
-        if level == 'module_pxd':
+        if ctx.level == 'module_pxd':
             body_level = 'c_class_pxd'
         else:
             body_level = 'c_class'
-        doc, body = p_suite(s, body_level, with_doc = 1)
+        doc, body = p_suite(s, Ctx(level = body_level), with_doc = 1)
     else:
         s.expect_newline("Syntax error in C class definition")
         doc = None
         body = None
-    if visibility == 'extern':
+    if ctx.visibility == 'extern':
         if not module_path:
             error(pos, "Module name required for 'extern' C class")
         if typeobj_name:
             error(pos, "Type object name specification not allowed for 'extern' C class")
-    elif visibility == 'public':
+    elif ctx.visibility == 'public':
         if not objstruct_name:
             error(pos, "Object struct name specification required for 'public' C class")
         if not typeobj_name:
             error(pos, "Type object name specification required for 'public' C class")
     else:
-        if api:
+        if ctx.api:
             error(pos, "Only 'public' C class can be declared 'api'")
     return Nodes.CClassDefNode(pos,
-        visibility = visibility,
-        typedef_flag = typedef_flag,
-        api = api,
+        visibility = ctx.visibility,
+        typedef_flag = ctx.typedef_flag,
+        api = ctx.api,
         module_name = ".".join(module_path),
         class_name = class_name,
         as_name = as_name,
@@ -1994,7 +1994,7 @@ def p_c_class_definition(s, level, pos,
         base_class_name = base_class_name,
         objstruct_name = objstruct_name,
         typeobj_name = typeobj_name,
-        in_pxd = level == 'module_pxd',
+        in_pxd = ctx.level == 'module_pxd',
         doc = doc,
         body = body)
 
@@ -2021,7 +2021,7 @@ def p_property_decl(s):
     pos = s.position()
     s.next() # 'property'
     name = p_ident(s)
-    doc, body = p_suite(s, 'property', with_doc = 1)
+    doc, body = p_suite(s, Ctx(level = 'property'), with_doc = 1)
     return Nodes.PropertyNode(pos, name = name, doc = doc, body = body)
 
 def p_doc_string(s):
@@ -2041,7 +2041,7 @@ def p_module(s, pxd):
         level = 'module_pxd'
     else:
         level = 'module'
-    body = p_statement_list(s, level)
+    body = p_statement_list(s, Ctx(level = level))
     if s.sy <> 'EOF':
         s.error("Syntax error in statement [%s,%s]" % (
             repr(s.sy), repr(s.systring)))
